//
// AHBulk.cpp
//
// $Revision: 1.8 $ $Date: 2007/10/19 18:12:52 $
//
// Copyright (C) 2007 Robert J. McDonald.  All rights reserved.
// http://www.dancingkayak.com/codesamples/anagrams/
//
// Loads all dictionary words in bulk into memory in one allocation and read.
// Doesn't create individual DictionaryEntry objects for each dictionary entry.
//

#include <fstream>

#include "AHBulk.h"

AHBulk::AHBulk() : m_wordbuffer(NULL), m_wordbufferLength(0), m_wordlist(NULL), m_freqTables(NULL), m_wordcount(0),
	m_entryCount(0), m_entriesLength(0), m_currentIndex(0), m_huntStatus(HUNT_NOT_STARTED)
{
}

AHBulk::~AHBulk()
{
	if (m_wordbuffer != NULL)
	{
		::free(m_wordbuffer);
		m_wordbuffer = NULL;
	}
	if (m_wordlist != NULL)
	{
		::free(m_wordlist);
		m_wordlist = NULL;
	}
	if (m_freqTables != NULL)
	{
		::free(m_freqTables);
		m_freqTables = NULL;
	}
}

// Specify the word for which anagrams will be searched
void AHBulk::setCandidate(const char * candidate)
{
	if (candidate == NULL)
	{
		m_dictEntry.setEntry("");
		m_errorMessage = "candidate is NULL";
	}
	else
	{
		m_dictEntry.setEntry(candidate);
		m_errorMessage = "";
	}
	m_huntStatus = HUNT_STARTED;
	m_currentIndex = 0;
	m_anagramCount = 0;
}

// Find first/next anagram of the candidate
const char * AHBulk::getAnagram(void)
{
	const char * nextWord = NULL;

	if (m_currentIndex < m_entryCount)
	{
		switch (m_huntStatus)
		{
		  case HUNT_NOT_STARTED:
			return NULL;
			break;

		  case HUNT_STARTED:
			m_huntStatus = HUNT_PROCEEDING;
			break;

		  case HUNT_PROCEEDING:
			m_currentIndex++;
			break;
		}

		while (m_currentIndex < m_entryCount)
		{
			if (0 != ::strcasecmp(m_wordlist[m_currentIndex].text, m_dictEntry.word()) &&
				0 == m_dictEntry.matchTable((int (*)[LETTER_COUNT])&m_freqTables[m_currentIndex * LETTER_COUNT]))
			{
				nextWord = m_wordlist[m_currentIndex].text;
				m_anagramCount++;
				break;
			}
			m_currentIndex++;
		}
	}
	return nextWord;
}

unsigned int AHBulk::readDictFile(const char * dictPath)
{
	unsigned int	returnCode;
	long		filelength;
	std::ifstream	dictionary;
	std::string	fileLine;

	returnCode = 0;

	dictionary.open(dictPath, std::ios::in);

	if (dictionary.is_open())
	{
		dictionary.exceptions(std::ifstream::failbit | std::ifstream::badbit);

		try
		{
			// Determine file size
			dictionary.seekg(0, std::ios::end);

			// Allocate buffer for file contents
			filelength = dictionary.tellg();
			m_wordbuffer = (char *) malloc(filelength);
			if (m_wordbuffer == NULL)
			{
				m_errorMessage = "";
				m_errorMessage += "Can't allocate buffer for file contents";
				returnCode = 2;
			}
			else
			{
				m_wordbufferLength = filelength;

				// Rewind file to beginning
				dictionary.seekg(0, std::ios::beg);

				// Read file contents into buffer
				dictionary.read(m_wordbuffer, m_wordbufferLength);
			}
		}
		catch (std::ifstream::failure e)
		{
			m_errorMessage = "";
			m_errorMessage += "Can't read file ";
			m_errorMessage += dictPath;
			m_errorMessage += " - ";
			m_errorMessage += e.what();
			returnCode = 3;
		}
		dictionary.close();
	}
	else
	{
		m_errorMessage = "";
		m_errorMessage += "Can't open file ";
		m_errorMessage += dictPath;
		returnCode = 1;
	}

	return returnCode;

}

unsigned int AHBulk::countEntries(void)
{
	unsigned int	returnCode;
	unsigned int	charIndex;
	int		inEntry;
	unsigned int	entryCount;

	returnCode = 0;

	entryCount = 0;
	inEntry = 0;

	// Count words
	for (charIndex = 0; charIndex < m_wordbufferLength; charIndex++)
	{
		if (m_wordbuffer[charIndex] == '\n')
		{
			inEntry = 0;
		}
		else
		{
			// Must be at least one non-newline to make a valid entry
			if (inEntry == 0)
			{
				// Use the beginning of the entry to boost count, since maybe no new line at EOF
				inEntry = 1;
				entryCount++;
			}
		}
	}

	m_entryCount = entryCount;

	// Allocate word pointer array
	m_wordlist = (struct wordInfo *) ::calloc(m_entryCount, sizeof(struct wordInfo));
	if (m_wordlist == NULL)
	{
		m_errorMessage = "";
		m_errorMessage += "Can't allocate word pointer array";
		returnCode = 4;
	}
	else
	{
		// Allocate frequency table array
		m_freqTables = (int *) ::calloc(m_entryCount * LETTER_COUNT, sizeof(int));
		if (m_freqTables == NULL)
		{
			m_errorMessage = "";
			m_errorMessage += "Can't allocate frequency table array";
			returnCode = 5;
		}
	}

	return returnCode;

}

unsigned int AHBulk::convertDict(void)
{
	unsigned int	returnCode;
	int		inEntry;
	unsigned int	entryIndex;
	unsigned int	charIndex;
	char*		wordStart;

	returnCode = 0;

	// Walk through file contents,
	// changing newlines into nulls,
	// recording words and performing frequency analysis as find newlines.

	inEntry = 0;
	entryIndex = 0;
	wordStart = NULL;

	for (charIndex = 0; charIndex < m_wordbufferLength; charIndex++)
	{
		if (m_wordbuffer[charIndex] == '\n')
		{
			if (inEntry)
			{
				// Change newline to null byte
				m_wordbuffer[charIndex] = '\0';
				// Complete the entry, record the word.
				m_wordlist[entryIndex].text = wordStart;
				m_wordlist[entryIndex].length = m_wordbuffer + charIndex - wordStart;
				entryIndex++;
				inEntry = 0;
			}
		}
		else
		{
			// Must be at least one non-newline to make a valid entry
			if (inEntry == 0)
			{
				// Use the beginning of the entry to boost count, since maybe no new line at EOF
				inEntry = 1;
				wordStart = m_wordbuffer + charIndex;
			}
		}
	}
	if (inEntry)
	{
		// Complete word terminated by end of buffer
		// Change newline to null byte
		m_wordbuffer[charIndex] = '\0';
		// Complete the entry, record the word.
		m_wordlist[entryIndex].text = wordStart;
		m_wordlist[entryIndex].length = m_wordbuffer + charIndex - wordStart;
		entryIndex++;
		inEntry = 0;
	}
	if (entryIndex != m_entryCount)
	{
		m_errorMessage = "";
		m_errorMessage += "Can't obtain consistent entry count (";
		m_errorMessage += entryIndex;
		m_errorMessage += ", ";
		m_errorMessage += m_entryCount;
		m_errorMessage += ")";
		returnCode = 4;
	}

	return returnCode;

}

unsigned int AHBulk::setDict(const char * dictPath)
{
	unsigned int	entryIndex;
	unsigned int	returnCode;

	m_anagramCount = 0;

	returnCode = readDictFile(dictPath);

	if (returnCode == 0)
	{
		returnCode = countEntries();
	}

	if (returnCode == 0)
	{
		returnCode = convertDict();
	}

	if (returnCode == 0)
	{
		// Perform frequency analysis
		int*	freqTable;
		for (entryIndex = 0, freqTable = m_freqTables; entryIndex < m_entryCount; entryIndex++, freqTable += LETTER_COUNT)
		{
			unsigned int	letterIndex;

			for (letterIndex = 0; letterIndex < m_wordlist[entryIndex].length; letterIndex++)
			{
				char testLetter;

				testLetter = tolower(m_wordlist[entryIndex].text[letterIndex]);
				freqTable[testLetter - 'a']++;
			}
		}
	}
	return returnCode;
}

