//
// Unit testing for anagram hunter
//
// Robert McDonald
//
// $Revision: 1.4 $ $Date: 2007/10/19 18:12:52 $
//
// Copyright (C) 2007 Robert J. McDonald.  All rights reserved.
// http://www.dancingkayak.com/codesamples/anagrams/
//

#include <errno.h>

#include <cppunit/TestFixture.h>
#include <cppunit/TestCaller.h>
#include <cppunit/ui/text/TestRunner.h>

#include "AHSingles.h"
#include "AHBulk.h"

static const char * SINGLES_NAME = "AHSingles";
static const char * BULK_NAME = "AHBulk";

static const char * EMPTY_DICT = "emptyDict";
static const char * DUMMY_CANDIDATE = "dummy";
static const char * ONE_DICT = "oneDict";
static const char * ONE_CANDIDATE = "a";
static const char * ONE_LIST[1] = { ONE_CANDIDATE };
static const char * OOT_DICT = "ootDict";
static const char * OOT_CANDIDATE = "ab";
static const char * OOT_LIST[1] = { "ba" };
static const char * TWO_DICT = "twoDict";
static const char * TWO_CANDIDATE = "xyz";
static const char * TWO_LIST[2] = { "zyx", "xzy" };
static const char * NOTIN_DICT = "notInDict";
static const char * NOTIN_CANDIDATE = "notinthere";
static const char * BAD_CANDIDATE = "*()";
static const char * NONEXISTENT_DICT = "DictionaryFileDoesNotExist";

enum DictStatus
{
	DS_EXISTS = 0,
	DS_DOES_NOT_EXIST = 1
};

struct CanonData
{
	DictStatus	dictStatus;
	const char *	dictFile;
	const char *	candidate;
	unsigned int	anagramCount;
	const char **	anagramList;	// Pointer to array of pointers to const char
};

static const struct CanonData nomatch =
{
	DS_EXISTS, EMPTY_DICT, DUMMY_CANDIDATE, 0, NULL
};

static const struct CanonData matchonlyone =
{
	DS_EXISTS, ONE_DICT, ONE_CANDIDATE, 0, NULL
};

static const struct CanonData matchOneOfTwo =
{
	DS_EXISTS, OOT_DICT, OOT_CANDIDATE, sizeof(OOT_LIST) / sizeof(const char *), OOT_LIST
};

static const struct CanonData matchTwo =
{
	DS_EXISTS, TWO_DICT, TWO_CANDIDATE, sizeof(TWO_LIST) / sizeof(const char *), TWO_LIST
};

static const struct CanonData notin =
{
	DS_EXISTS, NOTIN_DICT, NOTIN_CANDIDATE, 0, NULL
};

static const struct CanonData badCandidate =
{
	DS_EXISTS, TWO_DICT, BAD_CANDIDATE, 0, NULL
};

static const struct CanonData nonexistentDict =
{
	DS_DOES_NOT_EXIST, NONEXISTENT_DICT, DUMMY_CANDIDATE, 0, NULL
};

class AnagramHunterTest : public CppUnit::TestFixture
{
private:
	AnagramHunter*	m_singlesHunter;
	AnagramHunter*	m_bulkHunter;

	void runAnagramTest(const char * anagramHunterName, const struct CanonData & canonData, AnagramHunter & hunter);

public:
	AnagramHunterTest(void);

	// Virtual methods
	void setUp();
	void tearDown();

	void testNoMatch(void);
	void testMatchOnlyOne(void);
	void testMatchOneOfTwo(void);
	void testMatchTwo(void);
	void testNotIn(void);
	void testBadCandidate(void);
	void testNonexistentDict(void);

	static CppUnit::Test *suite(void);

};

void AnagramHunterTest::setUp()
{
	m_singlesHunter = new AHSingles();
	m_bulkHunter = new AHBulk();
}

void AnagramHunterTest::tearDown()
{
	if (m_singlesHunter != NULL)
	{
		delete m_singlesHunter;
	}
	if (m_bulkHunter != NULL)
	{
		delete m_bulkHunter;
	}
}

AnagramHunterTest::AnagramHunterTest()
{
	m_singlesHunter = NULL;
	m_bulkHunter = NULL;
}

void AnagramHunterTest::runAnagramTest(const char * anagramHunterName, const struct CanonData & canonData, AnagramHunter & hunter)
{
	const char *	currentAnagram;
	int		result;
	std::string	errorMessage;

	printf("\tTesting candidate '%s' dictionary file '%s' on AnagramTest '%s'\n", canonData.candidate, canonData.dictFile, anagramHunterName);

	result = hunter.setDict(canonData.dictFile);
	if (canonData.dictStatus == DS_EXISTS)
	{
		errorMessage = "Error setting dictionary:  ";
		errorMessage += hunter.errorMessage();
		CPPUNIT_ASSERT_MESSAGE(errorMessage.c_str(), result == 0);
	}
	else if (canonData.dictStatus == DS_DOES_NOT_EXIST)
	{
		errorMessage = "Unexpected success from setDict()";
		CPPUNIT_ASSERT_MESSAGE(errorMessage.c_str(), result != 0);
	}
	else
	{
		errorMessage = "Invalid value ";
		errorMessage += canonData.dictStatus;
		errorMessage += " for canonData.dictStatus";
		CPPUNIT_FAIL(errorMessage.c_str());
	}
	hunter.setCandidate(canonData.candidate);
	while (hunter.anagramCount() < canonData.anagramCount)
	{
		currentAnagram = hunter.getAnagram();
		errorMessage = "Test FAILED:  expected anagram '";
		errorMessage += canonData.anagramList[hunter.anagramCount()];
		errorMessage += "' not found";
		CPPUNIT_ASSERT_MESSAGE(errorMessage.c_str(), currentAnagram != NULL);

		// Compare actual anagram with expected anagram
		errorMessage = "Found anagram '";
		errorMessage += currentAnagram;
		errorMessage += "' expected anagram '";
		errorMessage += canonData.anagramList[hunter.anagramCount() - 1];
		CPPUNIT_ASSERT_MESSAGE(errorMessage.c_str(), 0 == ::strcasecmp(currentAnagram, canonData.anagramList[hunter.anagramCount() - 1]));
	}

	errorMessage = "Actual anagram count ";
	errorMessage += hunter.anagramCount();
	errorMessage += " differs from expected count ";
	errorMessage += canonData.anagramCount;
	CPPUNIT_ASSERT_MESSAGE(errorMessage.c_str(), hunter.anagramCount() == canonData.anagramCount);

	// Try one more time in case there are more anagrams than expected
	currentAnagram = hunter.getAnagram();
	errorMessage = "Unexpected extra anagram '";
	if (currentAnagram != NULL)
	{
		errorMessage += currentAnagram;
	}
	errorMessage += "' found";
	CPPUNIT_ASSERT_MESSAGE(errorMessage.c_str(), currentAnagram == NULL);
}

void AnagramHunterTest::testNoMatch(void)
{
	runAnagramTest(SINGLES_NAME, nomatch, *m_singlesHunter);
	runAnagramTest(BULK_NAME, nomatch, *m_bulkHunter);
}

void AnagramHunterTest::testMatchOnlyOne(void)
{
	runAnagramTest(SINGLES_NAME, matchonlyone, *m_singlesHunter);
	runAnagramTest(BULK_NAME, matchonlyone, *m_bulkHunter);
}

void AnagramHunterTest::testMatchOneOfTwo(void)
{
	runAnagramTest(SINGLES_NAME, matchOneOfTwo, *m_singlesHunter);
	runAnagramTest(BULK_NAME, matchOneOfTwo, *m_bulkHunter);
}

void AnagramHunterTest::testMatchTwo(void)
{
	runAnagramTest(SINGLES_NAME, matchTwo, *m_singlesHunter);
	runAnagramTest(BULK_NAME, matchTwo, *m_bulkHunter);
}

void AnagramHunterTest::testNotIn(void)
{
	runAnagramTest(SINGLES_NAME, notin, *m_singlesHunter);
	runAnagramTest(BULK_NAME, notin, *m_bulkHunter);
}

void AnagramHunterTest::testBadCandidate(void)
{
	runAnagramTest(SINGLES_NAME, badCandidate, *m_singlesHunter);
	runAnagramTest(BULK_NAME, badCandidate, *m_bulkHunter);
}

void AnagramHunterTest::testNonexistentDict(void)
{
	runAnagramTest(SINGLES_NAME, nonexistentDict, *m_singlesHunter);
	runAnagramTest(BULK_NAME, nonexistentDict, *m_bulkHunter);
}

CppUnit::Test *AnagramHunterTest::suite(void)
{
	CppUnit::TestSuite *	suiteOfTests = new CppUnit::TestSuite("AnagramHunterTest");

	suiteOfTests->addTest(new CppUnit::TestCaller<AnagramHunterTest>("testNoMatch", &AnagramHunterTest::testNoMatch));
	suiteOfTests->addTest(new CppUnit::TestCaller<AnagramHunterTest>("testMatchOnlyOne", &AnagramHunterTest::testMatchOnlyOne));
	suiteOfTests->addTest(new CppUnit::TestCaller<AnagramHunterTest>("testMatchOneOfTwo", &AnagramHunterTest::testMatchOneOfTwo));
	suiteOfTests->addTest(new CppUnit::TestCaller<AnagramHunterTest>("testMatchTwo", &AnagramHunterTest::testMatchTwo));
	suiteOfTests->addTest(new CppUnit::TestCaller<AnagramHunterTest>("testNotIn", &AnagramHunterTest::testNotIn));
	suiteOfTests->addTest(new CppUnit::TestCaller<AnagramHunterTest>("testBadCandidate", &AnagramHunterTest::testBadCandidate));
	suiteOfTests->addTest(new CppUnit::TestCaller<AnagramHunterTest>("testNonexistentDict", &AnagramHunterTest::testNonexistentDict));
	return suiteOfTests;
}

int main(int argc, char ** argv)
{
	CppUnit::TextUi::TestRunner runner;

	runner.addTest(AnagramHunterTest::suite());
	runner.run();
	return 0;
}

